# Copyright (C) 2008-2011 W. Trevor King <wking@drexel.edu>
#
# This file is part of pypiezo.
#
# pypiezo is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# pypiezo is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pypiezo.  If not, see <http://www.gnu.org/licenses/>.

"Control of a piezo-based atomic force microscope."

import numpy as _numpy

try:
    import matplotlib as _matplotlib
    import matplotlib.pyplot as _matplotlib_pyplot
    import time as _time  # for timestamping lines on plots
except (ImportError, RuntimeError), e:
    _matplotlib = None
    _matplotlib_import_error = e

from curses_check_for_keypress import CheckForKeypress as _CheckForKeypress

from . import LOG as _LOG
from . import base as _base
from . import package_config as _package_config
from . import surface as _surface


class AFMPiezo (_base.Piezo):
    """A piezo-controlled atomic force microscope.

    This particular class expects a single input channel for measuring
    deflection.  Other subclasses provide support for multi-segment
    deflection measurements.

    >>> from pprint import pprint
    >>> from pycomedi.device import Device
    >>> from pycomedi.subdevice import StreamingSubdevice
    >>> from pycomedi.channel import AnalogChannel
    >>> from pycomedi.constant import AREF, SUBDEVICE_TYPE, UNIT
    >>> from . import config
    >>> from . import surface

    >>> d = Device('/dev/comedi0')
    >>> d.open()

    >>> s_in = d.find_subdevice_by_type(SUBDEVICE_TYPE.ai,
    ...     factory=StreamingSubdevice)
    >>> s_out = d.find_subdevice_by_type(SUBDEVICE_TYPE.ao,
    ...     factory=StreamingSubdevice)

    >>> axis_channel = s_out.channel(
    ...     0, factory=AnalogChannel, aref=AREF.ground)
    >>> input_channel = s_in.channel(0, factory=AnalogChannel, aref=AREF.diff)
    >>> for chan in [axis_channel, input_channel]:
    ...     chan.range = chan.find_range(unit=UNIT.volt, min=-10, max=10)

    We set the minimum voltage for the `z` axis to -9 (a volt above
    the minimum possible voltage) to help with testing
    `.get_surface_position`.  Without this minimum voltage, small
    calibration errors could lead to a railed -10 V input for the
    first few surface approaching steps, which could lead to an
    `EdgeKink` error instead of a `FlatFit` error.

    >>> axis_config = config.AxisConfig()
    >>> axis_config.update(
    ...     {'gain':20, 'sensitivity':8e-9, 'minimum':-9})
    >>> axis_config['channel'] = config.OutputChannelConfig()
    >>> axis_config['channel']['name'] = 'z'
    >>> input_config = config.InputChannelConfig()
    >>> input_config['name'] = 'deflection'

    >>> a = _base.PiezoAxis(config=axis_config, axis_channel=axis_channel)
    >>> a.setup_config()

    >>> c = _base.InputChannel(config=input_config, channel=input_channel)
    >>> c.setup_config()

    >>> p = AFMPiezo(axes=[a], inputs=[c], name='Molly')

    >>> deflection = p.read_deflection()
    >>> deflection  # doctest: +SKIP
    34494L
    >>> p.deflection_dtype()
    <type 'numpy.uint16'>

    We need to know where we are before we can move somewhere
    smoothly.

    >>> pos = _base.convert_volts_to_bits(p.config.select_config(
    ...     'axes', 'z', get_attribute=_base.get_axis_name)['channel'], 0)
    >>> p.jump('z', pos)

    Usually `.move_to_pos_or_def` is used to approach the surface, but
    for testing we assume the z output channel is connected directly
    into the deflection input channel.

    >>> target_pos = _base.convert_volts_to_bits(
    ...     p.config.select_config('axes', 'channel',
            get_attribute=_base.get_axis_name), 2)
    >>> step = int((target_pos - pos)/5)
    >>> target_def = _base.convert_volts_to_bits(
    ...     p.config.select_config('inputs', 'deflection'), 3)
    >>> data = p.move_to_pos_or_def('z', target_pos, target_def, step=step,
    ...     return_data=True)
    >>> p.last_output == {'z': int(target_pos)}
    True
    >>> pprint(data)  # doctest: +SKIP
    {'deflection':
       array([32655, 33967, 35280, 36593, 37905, 39218, 39222], dtype=uint16),
     'z':
       array([32767, 34077, 35387, 36697, 38007, 39317, 39321], dtype=uint16)}

    That was a working position-limited approach.  Now move back to
    the center and try a deflection-limited approach.

    >>> p.jump('z', pos)
    >>> target_def = _base.convert_volts_to_bits(
    ...     p.config.select_config('inputs', 'deflection'), 1)
    >>> data = p.move_to_pos_or_def('z', target_pos, target_def, step=step,
    ...     return_data=True)
    >>> print (p.last_output['z'] < int(target_pos))
    True
    >>> pprint(data)  # doctest: +SKIP
    {'deflection': array([32655, 33968, 35281, 36593], dtype=uint16),
     'z': array([32767, 34077, 35387, 36697], dtype=uint16)}

    >>> p.wiggle_for_interference('z', offset=p.last_output['z'],
    ...     laser_wavelength=650e-9, keypress_test_mode=True)
    Press any key to continue

    >>> try:
    ...     p.get_surface_position('z', max_deflection=target_def)
    ... except surface.FlatFit, e:
    ...     print 'got FlatFit'
    got FlatFit
    >>> print e  # doctest: +SKIP
    slopes not sufficiently different: 1.0021 and 1.0021
    >>> abs(e.right_slope-1) < 0.1
    True
    >>> abs(e.left_slope-1) < 0.1
    True

    >>> d.close()
    """
    def _deflection_channel(self):
        return self.channel_by_name(name='deflection', direction='input')

    def read_deflection(self):
        """Return sensor deflection in bits.

        TODO: explain how bit <-> volt conversion will work for this
        "virtual" channel.
        """
        return self._deflection_channel().data_read()

    def deflection_dtype(self):
        "Return a Numpy dtype suitable for deflection bit values."
        return self._deflection_channel().subdevice.get_dtype()

    def move_to_pos_or_def(self, axis_name, position, deflection, step,
                           return_data=False, pre_move_steps=0):
        """TODO

        pre_move_steps : int
            number of 'null' steps to take before moving (confirming a
            stable input deflection).
        """
        if return_data or _package_config['matplotlib']:
            aquire_data = True
        else:
            aquire_data = False

        if step == 0:
            raise ValueError('must have non-zero step size')
        elif step < 0 and position > self.last_output[axis_name]:
            step = -step
        elif step > 0 and position < self.last_output[axis_name]:
            step = -step

        log_string = (
            'move to position %d or deflection %g on axis %s in steps of %d'
            % (position, deflection, axis_name, step))
        _LOG.debug(log_string)
        current_deflection = self.read_deflection()
        log_string = 'current position %d and deflection %g' % (
            self.last_output[axis_name], current_deflection)
        _LOG.debug(log_string)

        if aquire_data:
            def_array=[current_deflection]
            pos_array=[self.last_output[axis_name]]
        for i in range(pre_move_steps):
            self.jump(axis_name, piezo.last_output[axis_name])
            delection = self.read_deflection()
            if aquire_data:
                def_array.append(current_deflection)
                pos_array.append(self.last_output[axis_name])
        # step in until we hit our target position or exceed our target deflection
        while (self.last_output[axis_name] != position and
               current_deflection < deflection):
            dist_to = position - self.last_output[axis_name]
            if abs(dist_to) < abs(step):
                jump_to = position
            else:
                jump_to = self.last_output[axis_name] + step
            self.jump(axis_name, jump_to)
            current_deflection = self.read_deflection()
            log_string = (
                'current z piezo position %6d, current deflection %6d'
                % (current_deflection, self.last_output[axis_name]))
            _LOG.debug(log_string)
            if aquire_data:
                def_array.append(current_deflection)
                pos_array.append(self.last_output[axis_name])

        log_string = (
            'move to position %d or deflection %g on axis %s complete'
            % (position, deflection, axis_name))
        _LOG.debug(log_string)
        log_string = 'current position %d and deflection %g' % (
            self.last_output[axis_name], current_deflection)
        _LOG.debug(log_string)
        if _package_config['matplotlib']:
            if not _matplotlib:
                raise _matplotlib_import_error
            figure = _matplotlib_pyplot.figure()
            axes = figure.add_subplot(1, 1, 1)
            axes.hold(True)
            timestamp = _time.strftime('%H%M%S')
            axes.set_title('step approach %s' % timestamp)
            axes.plot(pos_array, def_array, '.', label=timestamp)
            #_pylab.legend(loc='best')
            figure.show()

        if return_data:
            data = {
                axis_name:_numpy.array(
                    pos_array, dtype=self.channel_dtype(
                        axis_name, direction='output')),
                'deflection':_numpy.array(
                    def_array, dtype=self.deflection_dtype()),
                }
            return data

    def wiggle_for_interference(
        self, axis_name, wiggle_frequency=2, n_samples=1024, amplitude=None,
        offset=None, laser_wavelength=None, plot=True,
        keypress_test_mode=False):
        """Output a sine wave and measure interference.

        With a poorly focused or aligned laser, leaked laser light
        reflecting off the surface may interfere with the light
        reflected off the cantilever, causing distance-dependent
        interference with a period roughly half the laser's
        wavelength.  This method wiggles the cantilever near the
        surface and monitors the magnitude of deflection oscillation,
        allowing the operator to adjust the laser alignment in order
        to minimize the interference.

        Modern commercial AFMs with computer-aligned lasers must do
        something like this automatically.
        """
        if _package_config['matplotlib']:
            plot = True
        if laser_wavelength and amplitude:
            log_string = \
                'use either laser_wavelength or amplitude, but not both'
            _LOG.warn(log_string)

        if None in (amplitude, offset):
            output_axis = self.axis_by_name(axis_name)
            maxdata = output_axis.axis_channel.get_maxdata()
            midpoint = int(maxdata/2)
            if offset == None:
                offset = midpoint
                log_string = (
                    'generated offset for interference wiggle: %g' % offset)
                _LOG.debug(log_string)
            if amplitude == None:
                if offset <= midpoint:
                    max_amplitude = int(offset)
                else:
                    max_amplitude = int(maxdata-offset)
                offset_meters = _base.convert_bits_to_meters(
                    output_axis.config, offset)
                bit_wavelength = _base.convert_meters_to_bits(
                    output_axis.config, offset_meters + laser_wavelength
                    ) - offset
                amplitude = 2*bit_wavelength
                log_string = (
                    'generated amplitude for interference wiggle: %g'
                    % amplitude)
                _LOG.debug(log_string)
                if amplitude > max_amplitude:
                    raise ValueError('no room for a two wavelength wiggle')

        scan_frequency = wiggle_frequency * n_samples
        out = (amplitude * _numpy.sin(
                _numpy.arange(n_samples) * 4 * _numpy.pi / float(n_samples))
               + offset)
        # 4 for 2 periods, so you can judge precision
        out = out.reshape((len(out), 1)).astype(
            self.channel_dtype(axis_name, direction='output'))

        _LOG.debug('oscillate for interference wiggle')
        log_string = (
            'amplitude: %d bits, offset: %d bits, scan frequency: %g Hz'
            % (amplitude, offset, scan_frequency))
        _LOG.debug(log_string)

        if plot:
            if not _matplotlib:
                raise _matplotlib_import_error
            figure = _matplotlib_pyplot.figure()
            axes = figure.add_subplot(1, 1, 1)
            axes.hold(False)
            timestamp = _time.strftime('%H%M%S')
            axes.set_title('wiggle for interference %s' % timestamp)
            plot_p = axes.plot(out, out, 'b.-')
            figure.show()
        c = _CheckForKeypress(test_mode=keypress_test_mode)
        while c.input() == None:
            # input will need processing for multi-segment AFMs...
            data = self.ramp(out, scan_frequency, output_names=[axis_name],
                             input_names=['deflection'])
            _LOG.debug('completed a wiggle round')
            if plot:
                plot_p[0].set_ydata(data[:,0])
                axes.set_ylim([data.min(), data.max()])
                #_flush_plot()
        self.last_output[axis_name] = out[-1,0]
        _LOG.debug('interference wiggle complete')

    get_surface_position = _surface.get_surface_position


#def ramp
#        if USE_ABCD_DEFLECTION :
#            for i in range(4) : # i is the photodiode element (0->A, 1->B, ...)
#                self.curIn[i] = out["Deflection segment"][i][-1]
#        else :
#            self.curIn[self.chan_info.def_ind] = out["deflection"][-1]


#class FourSegmentAFM (AFM):
#    def read_deflection(self):
#        "Return sensor deflection in bits."
#        A = int(self.curIn[self.chan_info.def_ind[0]])
#        B = int(self.curIn[self.chan_info.def_ind[1]])
#        C = int(self.curIn[self.chan_info.def_ind[2]])
#        D = int(self.curIn[self.chan_info.def_ind[3]])
#        df = float((A+B)-(C+D))/(A+B+C+D)
#        dfout = int(df * 2**15) + 2**15
#        if TEXT_VERBOSE :
#            print "Current deflection %d (%d, %d, %d, %d)" \
#                % (dfout, A, B, C, D)
#        return dfout


#def test_smoothness(zp, plotVerbose=True):
#    posA = 20000
#    posB = 50000
#    setpoint = zp.def_V2in(3)
#    steps = 200
#    outfreq = 1e5
#    outarray = linspace(posB, posA, 1000)
#    indata=[]
#    outdata=[]
#    curVals = zp.jumpToPos(posA)
#    zp.pCurVals(curVals)
#    _sleep(1) # let jitters die down
#    for i in range(10):
#        print "ramp %d to %d" % (zp.curPos(), posB)
#        curVals, data = moveToPosOrDef(zp, posB, setpoint, step=steps,
#                                       return_data = True)
#        indata.append(data)
#        out = zp.ramp(outarray, outfreq)
#        outdata.append(out)
#    if plotVerbose:
#        from pylab import figure, plot, title, legend, hold, subplot        
#    if PYLAB_VERBOSE or plotVerbose:
#        _import_pylab()
#        _pylab.figure(BASE_FIG_NUM+4)
#        for i in range(10):
#            _pylab.plot(indata[i]['z'],
#                        indata[i]['deflection'], '+--', label='in')
#            _pylab.plot(outdata[i]['z'],
#                        outdata[i]['deflection'], '.-', label='out')
#        _pylab.title('test smoothness (step in, ramp out)')
#        #_pylab.legend(loc='best')
#    
#def test():
#    import z_piezo
#    zp = z_piezo.z_piezo()
#    curVals = zp.moveToPosOrDef(zp.pos_nm2out(600), defl=zp.curDef()+6000, step=(zp.pos_nm2out(10)-zp.pos_nm2out(0)))
#    if TEXT_VERBOSE:
#        zp.pCurVals(curVals)
#    pos = zp.getSurfPos(maxDefl=zp.curDef()+6000)
#    if TEXT_VERBOSE:
#        print "Surface at %g nm", pos
#    print "success"
#    if PYLAB_VERBOSE and _final_flush_plot != None:
#        _final_flush_plot()

